#include <algorithm>
#include <cfloat>
#include <cstring>
#include <fstream>
#include <iostream>
#include <limits>
#include <ostream>
#include <random>
#include <stdexcept>
#include <vector>

#include <boost/filesystem.hpp>

#include <vulkan/vulkan.h>

namespace apgk
{

//-------------------------------------------------------------------------------------------------
// vkCreateDebugReportCallbackEXT
//

VkResult CreateDebugReportCallbackEXT(
        VkInstance                                  instance,
        VkDebugReportCallbackCreateInfoEXT const*   pCreateInfo,
        VkAllocationCallbacks const*                pAllocator,
        VkDebugReportCallbackEXT*                   pCallback
        )
{
    auto func = (PFN_vkCreateDebugReportCallbackEXT)vkGetInstanceProcAddr(instance, "vkCreateDebugReportCallbackEXT");
    if (func != nullptr)
    {
        return func(instance, pCreateInfo, pAllocator, pCallback);
    }
    else
    {
        return VK_ERROR_EXTENSION_NOT_PRESENT;
    }
}

//-------------------------------------------------------------------------------------------------
// vkDestroyDebugReportCallbackEXT
//

void DestroyDebugReportCallbackEXT(
        VkInstance                      instance,
        VkDebugReportCallbackEXT        callback,
        VkAllocationCallbacks const*    pAllocator
        )
{
    auto func = (PFN_vkDestroyDebugReportCallbackEXT)vkGetInstanceProcAddr(instance, "vkDestroyDebugReportCallbackEXT");
    if (func != nullptr)
    {
        func(instance, callback, pAllocator);
    }
}

} // apgk


void read_bytes(std::string filename, std::vector<char>& bytes)
{
    // Open file and start reading at the beginning to determine its size
    std::ifstream file(filename, std::ifstream::ate | std::ifstream::binary);

    bytes.resize(static_cast<std::size_t>(file.tellg()));

    file.seekg(0);

    file.read(bytes.data(), bytes.size());
}


//-------------------------------------------------------------------------------------------------
// Try to load vertex and fragment shaders
//

bool try_load(std::string filename, std::vector<char>& bytes)
{
    namespace fs = boost::filesystem;

    std::cout << "Try to load shader from:\n";
    std::cout << "  " << filename << '\n';
    if (fs::exists(filename))
    {
        read_bytes(filename, bytes);
        std::cout << "Success!\n";
        return true;
    }
    else
    {
        std::cout << "File does not exist..\n";
        return false;
    }
}


//-------------------------------------------------------------------------------------------------
// Load shader in shaders/ folder
//

bool load_shader(std::vector<char>& bytes)
{
    std::string filename = "saxpy.comp.spv";

    // Try to find the two shader files
    bool loaded = false;

    // Environment variable SHADER_DIR overrides!
    char* shader_dir = getenv("SHADER_DIR");
    if (!loaded && shader_dir)
    {   
        loaded = try_load(
            std::string(shader_dir) + '/' + filename,
            bytes
            );  
    }   

    if (!loaded)
    {   
        loaded = try_load("shaders/" + filename, bytes);
    }   

    if (!loaded)
    {   
        loaded = try_load(filename, bytes);
    }   

    return loaded;
}

bool checkPhysicalDevice(VkPhysicalDevice& dev, int& first_compute_queue)
{
    VkPhysicalDeviceProperties dprops;
    vkGetPhysicalDeviceProperties(dev, &dprops);
    std::cout << "Device Name   : " << dprops.deviceName << '\n';
    std::cout << "API Version   : ";
    std::cout << VK_VERSION_MAJOR(dprops.apiVersion) << '.'
        << VK_VERSION_MINOR(dprops.apiVersion) << '.'
        << VK_VERSION_PATCH(dprops.apiVersion) << '\n';
    std::cout << "Driver Version: ";
    std::cout << VK_VERSION_MAJOR(dprops.driverVersion) << '.'
        << VK_VERSION_MINOR(dprops.driverVersion) << '.'
        << VK_VERSION_PATCH(dprops.driverVersion) << '\n';
    // Implement lots of checks here to pick the right device..

    VkPhysicalDeviceFeatures dfeats;
    vkGetPhysicalDeviceFeatures(dev, &dfeats);

    uint32_t qfcount = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(dev, &qfcount, nullptr);

    std::vector<VkQueueFamilyProperties> queuefamilies(qfcount);
    vkGetPhysicalDeviceQueueFamilyProperties(dev, &qfcount, queuefamilies.data());

    first_compute_queue = -1;

    int i = 0;
    for (auto& fam : queuefamilies)
    {
        if (first_compute_queue < 0 && fam.queueFlags & VK_QUEUE_COMPUTE_BIT)
        {
            first_compute_queue = i;
        }
        ++i;
    }

    return first_compute_queue >= 0;
}

class saxpy_application
{
public:

    std::vector<char const*> required_instance_extensions()
    {
        return std::vector<char const*>({
            VK_EXT_DEBUG_REPORT_EXTENSION_NAME
            });
    };

    std::vector<char const*> required_device_extensions()
    {
        return std::vector<char const*>({
            VK_KHR_SWAPCHAIN_EXTENSION_NAME
            });
    };

    void create_buffer(VkBuffer& buffer, VkDeviceMemory& mem, std::size_t N)
    {
        VkBufferCreateInfo binfo = {};
        binfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
        binfo.size = static_cast<uint32_t>(sizeof(float) * N);
        binfo.usage = VK_BUFFER_USAGE_STORAGE_BUFFER_BIT;
        binfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

        if (vkCreateBuffer(device_, &binfo, nullptr, &buffer) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateBuffer failed");
        }

        VkMemoryRequirements req;
        vkGetBufferMemoryRequirements(device_, buffer, &req);

        VkMemoryAllocateInfo allocinfo = {};
        allocinfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
        allocinfo.allocationSize = req.size;

        uint32_t mti(-1);

        // Search for the right memory type index
        VkMemoryPropertyFlags req_properties = VK_MEMORY_PROPERTY_HOST_COHERENT_BIT
                                             | VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT;
        VkPhysicalDeviceMemoryProperties props;
        vkGetPhysicalDeviceMemoryProperties(physical_device_, &props);
        for (uint32_t i = 0; i < props.memoryTypeCount; ++i)
        {
            if ((req.memoryTypeBits & (1 << i)) && (props.memoryTypes[i].propertyFlags & req_properties) == req_properties)
            {
                mti = i;
                break;
            }
        }
        allocinfo.memoryTypeIndex = mti;

        // Allocate device memory
        if (vkAllocateMemory(device_, &allocinfo, nullptr, &mem) != VK_SUCCESS)
        {
            throw std::runtime_error("vkAllocateMemory() failed");
        }

        if (vkBindBufferMemory(device_, buffer, mem, 0) != VK_SUCCESS)
        {
            throw std::runtime_error("vkBindBufferMemory() failed");
        }
    }

    void fill_input_buffers()
    {
        // Map the device memory to host addres
        // space so we can copy to it
        void* vy = nullptr;
        vkMapMemory(device_, buffer_memory_[0], 0, VK_WHOLE_SIZE, 0, &vy);
        float* y = static_cast<float*>(vy);

        void* va = nullptr;
        vkMapMemory(device_, buffer_memory_[1], 0, VK_WHOLE_SIZE, 0, &va);
        float* a = static_cast<float*>(va);

        void* vx = nullptr;
        vkMapMemory(device_, buffer_memory_[2], 0, VK_WHOLE_SIZE, 0, &vx);
        float* x = static_cast<float*>(vx);

        std::copy(h_y.begin(), h_y.end(), y);
        std::copy(h_a.begin(), h_a.end(), a);
        std::copy(h_x.begin(), h_x.end(), x);

        // Unmap memory
        vkUnmapMemory(device_, buffer_memory_[2]);
        vkUnmapMemory(device_, buffer_memory_[1]);
        vkUnmapMemory(device_, buffer_memory_[0]);
    }

    void retrieve_output_buffer(std::vector<float>& output, std::size_t index = 0)
    {
        // Map the device memory to host addres
        // space so we can copy from it
        void* vy = nullptr;
        vkMapMemory(device_, buffer_memory_[index], 0, VK_WHOLE_SIZE, 0, &vy);
        float* y = static_cast<float*>(vy);

        output.resize(N);
        std::copy(y, y + N, output.data());

        // Unmap memory
        vkUnmapMemory(device_, buffer_memory_[index]);
    }

    void init(int argc, char** argv)
    {
        // Fill host buffers with random values
        std::default_random_engine rng;
        std::uniform_real_distribution<float> dist;

        h_y.resize(N);
        h_a.resize(N);
        h_x.resize(N);

        for (std::size_t i = 0; i < N; ++i)
        {
            h_y[i] = dist(rng);
            h_a[i] = dist(rng);
            h_x[i] = dist(rng);
        }


    //--- extensions
        uint32_t ext_count = 0;
        vkEnumerateInstanceExtensionProperties(nullptr, &ext_count, nullptr);

#ifdef VERBOSE
        std::cout << ext_count << '\n';
#endif
        std::vector<VkExtensionProperties> extensions(ext_count);
        vkEnumerateInstanceExtensionProperties(nullptr, &ext_count, extensions.data());

#ifdef VERBOSE
        for (auto const& ext : extensions)
        {
            std::cout << ext.extensionName << '\n';
        }
#endif

        std::vector<char const*> extensions_desired = required_instance_extensions();
        /* TODO: check if all required extensions are available */

    //--- validation layers
        std::vector<char const*> vlayers = {
            "VK_LAYER_LUNARG_standard_validation"
        };

        uint32_t layer_count = 0;
        vkEnumerateInstanceLayerProperties(&layer_count, nullptr);

#ifdef VERBOSE
        std::cout << layer_count << '\n';
#endif
        std::vector<VkLayerProperties> layers(layer_count);
        vkEnumerateInstanceLayerProperties(&layer_count, layers.data());

#ifdef VERBOSE
        for (auto const& layer : layers)
        {
            std::cout << layer.layerName << '\n';
        }
#endif

    //--- activate validation layers
        std::vector<char const*> layers_desired({
            "VK_LAYER_LUNARG_standard_validation"
            });

        for (auto ld : layers_desired)
        {
            bool found = false;

            for (auto l : layers)
            {
                if (strcmp(l.layerName, ld) == 0)
                {
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                throw std::runtime_error("Requested validation layer not found!");
            }
        }

    //--- instance
        VkApplicationInfo ainfo = {};
        ainfo.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
        ainfo.pApplicationName = "Vulkan Application";
        ainfo.applicationVersion = VK_MAKE_VERSION(0, 0, 1);
        ainfo.pEngineName = "";
        ainfo.engineVersion = VK_MAKE_VERSION(0, 0, 1);
        ainfo.apiVersion = VK_API_VERSION_1_1;

        VkInstanceCreateInfo cinfo = {};
        cinfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
        cinfo.pApplicationInfo = &ainfo;

        cinfo.enabledLayerCount = static_cast<uint32_t>(layers_desired.size());
        cinfo.ppEnabledLayerNames = layers_desired.data();

        cinfo.enabledExtensionCount = static_cast<uint32_t>(
            extensions_desired.size()
            );
        cinfo.ppEnabledExtensionNames = extensions_desired.data();

        if (vkCreateInstance(&cinfo, nullptr, &instance_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateInstance failed!");
        }


//--- debug callback

        VkDebugReportCallbackCreateInfoEXT cbinfo = {};
        cbinfo.sType = VK_STRUCTURE_TYPE_DEBUG_REPORT_CALLBACK_CREATE_INFO_EXT;
        cbinfo.flags = VK_DEBUG_REPORT_ERROR_BIT_EXT | VK_DEBUG_REPORT_WARNING_BIT_EXT;

        cbinfo.pfnCallback = [](
                VkDebugReportFlagsEXT flags,
                VkDebugReportObjectTypeEXT obj_type,
                uint64_t obj,
                std::size_t location,
                int32_t code,
                char const* layer_prefix,
                char const* message,
                void* user_data
                )
            -> VkBool32
        {
            std::cerr << message << '\n';

            return VK_FALSE;
        };

        if (apgk::CreateDebugReportCallbackEXT(instance_, &cbinfo, nullptr, &debug_callback_) != VK_SUCCESS)
        {
            throw std::runtime_error("apgk::CreateDebugReportCallbackEXT failed!");
        }

//--- pick physical device

        uint32_t dcount = 0;
        vkEnumeratePhysicalDevices(instance_, &dcount, nullptr);
#ifdef VERBOSE
        std::cout << dcount << '\n';
#endif

        if (dcount < 1)
        {
            throw std::runtime_error("No physical devices!");
        }

        std::vector<VkPhysicalDevice> devices(dcount);
        vkEnumeratePhysicalDevices(instance_, &dcount, devices.data());
        int first_suitable_device = -1;
        int first_compute_queue = -1;
        for (std::size_t i = 0; i < devices.size(); ++i)
        {
            std::cout << "Physical Device " << i << '\n';
            std::cout << "=================\n";
            if (checkPhysicalDevice(devices[i], first_compute_queue))
            {
                first_suitable_device = static_cast<int>(i);
            }
            std::cout << '\n';
        }

        if (first_suitable_device < 0)
        {
            throw std::runtime_error("No suitable physical device found");
        }

        if (first_compute_queue < 0)
        {
            throw std::runtime_error("No device with a compute queue found!");
        }

        // Just pick the first suitable device
        physical_device_ = devices[first_suitable_device];

//--- create logical device

        VkDeviceQueueCreateInfo qinfo = {};
        qinfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
        qinfo.queueFamilyIndex = first_compute_queue;
        qinfo.queueCount = 1;

        float queue_priority = 1.0f; // may go out of scope?
        qinfo.pQueuePriorities = &queue_priority;

        // Nothing special for now
        VkPhysicalDeviceFeatures device_features = {};

        // Device extensions
        extensions_desired = required_device_extensions();

        // Create the logical device
        VkDeviceCreateInfo dcinfo = {};
        dcinfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
        dcinfo.pQueueCreateInfos = &qinfo;
        dcinfo.queueCreateInfoCount = 1;
        dcinfo.pEnabledFeatures = &device_features;
        // Per-device validation layers
        dcinfo.enabledLayerCount = static_cast<uint32_t>(layers_desired.size());
        dcinfo.ppEnabledLayerNames = layers_desired.data();
        // Per-device extensions
        dcinfo.enabledExtensionCount = static_cast<uint32_t>(extensions_desired.size());
        dcinfo.ppEnabledExtensionNames = extensions_desired.data();
        // Actually create..
        if (vkCreateDevice(physical_device_, &dcinfo, nullptr, &device_) != VK_SUCCESS)
        {
            throw std::runtime_error("Could not create logical device");
        }

//--- Queue handles
        vkGetDeviceQueue(device_, first_compute_queue, 0, &compute_queue_);

//--- Input Buffers
        create_buffer(buffer_[0], buffer_memory_[0], N);
        create_buffer(buffer_[1], buffer_memory_[1], N);
        create_buffer(buffer_[2], buffer_memory_[2], N);


        fill_input_buffers();

//--- Descriptor set

        // First create descriptor set layouts to bind descriptor
        // resources in compute shader
        VkDescriptorSetLayoutBinding lbs[3];
        for (uint32_t i = 0; i < 3; ++i)
        {
            std::memset(&lbs[i], 0, sizeof(lbs[i]));
            lbs[i].binding = i;
            lbs[i].descriptorType = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
            lbs[i].descriptorCount = 1;
            lbs[i].stageFlags = VK_SHADER_STAGE_COMPUTE_BIT;
        }

        VkDescriptorSetLayoutCreateInfo dsinfo = {};
        dsinfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
        dsinfo.bindingCount = 3;
        dsinfo.pBindings = lbs;

        if (vkCreateDescriptorSetLayout(device_, &dsinfo, nullptr, &descriptor_set_layout_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateDescriptorSetLayout failed");
        }

        // We need a pool to allocate a descriptor set from
        VkDescriptorPoolSize dpsize = {};
        dpsize.type = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
        dpsize.descriptorCount = 1;

        VkDescriptorPoolCreateInfo dpinfo = {};
        dpinfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
        dpinfo.maxSets = 1;
        dpinfo.poolSizeCount = 1;
        dpinfo.pPoolSizes = &dpsize;

        if (vkCreateDescriptorPool(device_, &dpinfo, nullptr, &descriptor_pool_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateDescriptorPool failed");
        }

        // Allocate a descriptor set that bind the three buffers to
        VkDescriptorSetAllocateInfo dainfo = {};
        dainfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
        dainfo.descriptorPool = descriptor_pool_;
        dainfo.descriptorSetCount = 1;
        dainfo.pSetLayouts = &descriptor_set_layout_;

        if (vkAllocateDescriptorSets(device_, &dainfo, &descriptor_set_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkAllocateDescriptorSets failed");
        }

        VkWriteDescriptorSet wdsets[3];
        for (uint32_t i = 0; i < 3; ++i)
        {
            VkDescriptorBufferInfo buffer_info = {};
            buffer_info.buffer = buffer_[i];
            buffer_info.offset = 0;
            buffer_info.range = VK_WHOLE_SIZE;

            std::memset(&wdsets[i], 0, sizeof(wdsets[i]));
            wdsets[i].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            wdsets[i].dstSet = descriptor_set_;
            wdsets[i].dstBinding = i;
            wdsets[i].descriptorCount = 1;
            wdsets[i].descriptorType = VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
            wdsets[i].pBufferInfo = &buffer_info;
            vkUpdateDescriptorSets(device_, 1, &wdsets[i], 0, nullptr);
        }

//--- Compute pipeline

        // In contrast to a graphics pipeline, compute pipelines
        // only have a single (shader) stage

        std::vector<char> bytes;
        if (!load_shader(bytes))
        {
            throw std::runtime_error("Failed to load SPIR-V shader");
        }

        VkShaderModuleCreateInfo sminfo = {};
        sminfo.sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
        sminfo.pCode = reinterpret_cast<uint32_t const*>(bytes.data());
        sminfo.codeSize = static_cast<uint32_t>(bytes.size());

        if (vkCreateShaderModule(device_, &sminfo, nullptr, &saxpy_shader_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateShaderModule failed");
        }

        // In order to create a pipeline, we need a pipeline shader stage..
        VkPipelineShaderStageCreateInfo ssinfo = {};
        ssinfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
        ssinfo.stage = VK_SHADER_STAGE_COMPUTE_BIT;
        ssinfo.module = saxpy_shader_;
        ssinfo.pName = "main";

        // ..and a pipeline layout to later access descriptor sets.
        VkPipelineLayoutCreateInfo plinfo = {};
        plinfo.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
        plinfo.setLayoutCount = 1;
        plinfo.pSetLayouts = &descriptor_set_layout_;

        if (vkCreatePipelineLayout(device_, &plinfo, nullptr, &pipeline_layout_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreatePipelineLayout failed");
        }


        // Create compute pipeline
        VkComputePipelineCreateInfo cpinfo = {};
        cpinfo.sType = VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO;
        cpinfo.stage = ssinfo;
        cpinfo.layout = pipeline_layout_;

        if (vkCreateComputePipelines(device_,
                VK_NULL_HANDLE, 1, &cpinfo, nullptr, &pipeline_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateComputePipelines failed");
        }

//--- Command buffer

        // We need a pool to allocate the command buffer from
        VkCommandPoolCreateInfo poinfo = {};
        poinfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
        poinfo.flags = 0;
        poinfo.queueFamilyIndex = first_compute_queue;

        if (vkCreateCommandPool(device_, &poinfo, nullptr, &command_pool_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateCommandPool failed");
        }

        // Allocate the command buffer
        VkCommandBufferAllocateInfo alinfo = {};
        alinfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
        alinfo.commandPool = command_pool_;
        // Make a primary command buffer, can be directly submitted
        // to a queue
        alinfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
        alinfo.commandBufferCount = 1;

        if (vkAllocateCommandBuffers(device_, &alinfo, &command_buffer_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkAllocateCommandBuffers failed");
        }


        // Now start recording commands
        VkCommandBufferBeginInfo beginfo = {};
        beginfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        // This tells the application that we will run this command
        // buffer only once!
        beginfo.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

        if (vkBeginCommandBuffer(command_buffer_, &beginfo) != VK_SUCCESS)
        {
            throw std::runtime_error("vkBeginCommandBuffer failed");
        }

        // Commands (note how vkCmdXXX does not generate error codes!)
        vkCmdBindPipeline(command_buffer_, VK_PIPELINE_BIND_POINT_COMPUTE, pipeline_);
        vkCmdBindDescriptorSets(
                command_buffer_,
                VK_PIPELINE_BIND_POINT_COMPUTE,
                pipeline_layout_,
                0,
                1,
                &descriptor_set_,
                0,
                nullptr
                );


        // Go for it!
        uint32_t workgroup_size = 32; // TODO: should match with shader
        auto div_up = [](uint32_t a, uint32_t b)
        {
            return (a + b - 1) / b;
        };

        std::cout << "Dispatching " << div_up(N, workgroup_size) << " work group(s)\n";
        vkCmdDispatch(command_buffer_, div_up(N, workgroup_size), 1, 1);

        if (vkEndCommandBuffer(command_buffer_) != VK_SUCCESS)
        {
            throw std::runtime_error("vkEndCommandBuffer failed");
        }
    }

    void run()
    {
        // Finally go for it and submit the command buffer!
        VkSubmitInfo sinfo = {};
        sinfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
        sinfo.commandBufferCount = 1;
        sinfo.pCommandBuffers = &command_buffer_;

        VkFence fence;
        VkFenceCreateInfo finfo = {};
        finfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
        finfo.flags = 0;

        if (vkCreateFence(device_, &finfo, nullptr, &fence) != VK_SUCCESS)
        {
            throw std::runtime_error("vkCreateFence failed");
        }

        if (vkQueueSubmit(compute_queue_, 1, &sinfo, fence) != VK_SUCCESS)
        {
            throw std::runtime_error("vkQueueSubmit failed");
        }

        // Command buffer execution is an asynchronous operation
        // ==> wait using the fence/barrier
        if (vkWaitForFences(device_, 1, &fence, VK_TRUE,
                    std::numeric_limits<uint64_t>::max()) != VK_SUCCESS)
        {
            throw std::runtime_error("vkWaitForFences failed");
        }

        vkDestroyFence(device_, fence, nullptr);

        std::vector<float> output;
        retrieve_output_buffer(output, 0);

        // Perform saxpy operation on CPU and compare
        for (std::size_t i = 0; i < N; ++i)
        {
            h_y[i] = h_a[i] * h_x[i] + h_y[i];
            //std::cout << h_y[i] << ' ' << output[i] << '\n';
            if (std::abs(h_y[i] - output[i]) > FLT_EPSILON)
            {
                throw std::runtime_error("Error in SAXPY calculation");
            }
        }

        std::cout << "SAXPY calculation successfully finished!\n";
    }

    void cleanup()
    {
        // Free resources in the opposite order we created them

        // We don't need to free the command buffer back to the pool
        // because when the pool is destroyed the command buffer is
        // automatically freed.
        vkDestroyCommandPool(device_, command_pool_, nullptr);
        vkDestroyPipeline(device_, pipeline_, nullptr);
        vkDestroyPipelineLayout(device_, pipeline_layout_, nullptr);
        vkDestroyShaderModule(device_, saxpy_shader_, nullptr);
        // Same as for command pool allocation
        vkDestroyDescriptorPool(device_, descriptor_pool_, nullptr);
        vkDestroyDescriptorSetLayout(device_, descriptor_set_layout_, nullptr);
        for (int i = 0; i < 3; ++i)
        {
            vkFreeMemory(device_, buffer_memory_[i], nullptr);
            vkDestroyBuffer(device_, buffer_[i], nullptr);
        }
        vkDestroyDevice(device_, nullptr);
        apgk::DestroyDebugReportCallbackEXT(instance_, debug_callback_, nullptr);
        vkDestroyInstance(instance_, nullptr);
    }

private:

    // Instance, device, etc.
    VkInstance instance_;
    VkDebugReportCallbackEXT debug_callback_;
    VkPhysicalDevice physical_device_ = VK_NULL_HANDLE;
    VkDevice device_;
    VkQueue compute_queue_;

    // Compute pipeline
    VkPipeline pipeline_;
    VkPipelineLayout pipeline_layout_;
    VkShaderModule saxpy_shader_;
    VkDescriptorPool descriptor_pool_;
    VkDescriptorSet descriptor_set_;
    VkDescriptorSetLayout descriptor_set_layout_;

    VkCommandPool command_pool_;
    VkCommandBuffer command_buffer_;


    // Memory, buffers, etc.
    constexpr static std::size_t N = 1024 * 1024;

    VkBuffer buffer_[3];
    VkDeviceMemory buffer_memory_[3];

    // Host memory, copied to storage buffers
    std::vector<float> h_y;
    std::vector<float> h_a;
    std::vector<float> h_x;
};

int main(int argc, char** argv)
{
    saxpy_application app;
    app.init(argc, argv);
    app.run();
    app.cleanup();
}
